var commandLength = {  // #bits in each command
	'REG_WRITE_MASK_PAGE0_IMM': 28,
    'REG_WRITE_IMM': 22,		
	'SET_DELAY': 22,			// Compiles in to REG_WRITE_IMM statement
	'REG_WRITE_MASK_IMM': 30,	
	'REG_WRITE_VOUT_IMM': 24,
	'REG_WRITE_VCTRL_IMM': 32,	
	'REG_WRITE_MASK_SREG': 24,		
	'SREG_READ_REG': 16,				
	'WAIT': 24,
	'JUMP': 24,					// Compiles in to WAIT statement
	'DELAY_IMM': 12,			
	'DELAY_SREG': 6,					
	'TRIG_SET': 32,				
	'TRIG_MASK': 32,			
	'END': 4,
    'REG_WRITE_BIT_PAGE0_IMM': 16,
    'REG_WRITE_WIN_PAGE0_IMM': 24,
    'SREG_WRITE_IMM': 14
}

var opcodeMap = {
	'REG_WRITE_MASK_PAGE0_IMM': '0000',
    'REG_WRITE_IMM': 		    '0001',
	'SET_DELAY': 			    '0001',
	'REG_WRITE_MASK_IMM':	    '0010',
	'REG_WRITE_VOUT_IMM':	    '0011',
	'REG_WRITE_VCTRL_IMM':	    '0100',
	'REG_WRITE_MASK_SREG': 	    '0101',
	'SREG_READ_REG': 		    '0110',
	'WAIT': 				    '0111',
	'JUMP':					    '0111',
	'DELAY_IMM': 			    '1000',
	'DELAY_SREG':			    '1001',
	'TRIG_SET': 			    '1010',
	'TRIG_MASK':			    '1011',
	'END': 					    '1100',
    'REG_WRITE_BIT_PAGE0_IMM':  '1101',
    'REG_WRITE_WIN_PAGE0_IMM':  '1110',
    'SREG_WRITE_IMM':           '1111'
}

var emitEnd = function(params) {
    return opcodeMap['END'];
}

var emitRegWriteImm = function(params) {
    
    var addr = zeroPad(params.addr.toString(2), 10);
    var data = zeroPad(params.data.toString(2), 8);
    
    return data + addr + opcodeMap['REG_WRITE_IMM'];
}

var emitRegWriteMaskImm = function(params) {
    
    var addr = zeroPad(params.addr.toString(2), 10);
    var data = zeroPad(params.data.toString(2), 8);
    var mask = zeroPad(params.mask.toString(2), 8);
    
    return mask + data + addr + opcodeMap['REG_WRITE_MASK_IMM'];
}

var emitRegWriteVoutImm = function(params) {
    
    var regulator = zeroPad(params.regulator.toString(2), 4);
    var sel = zeroPad(params.sel.toString(2), 2);
    var vset = zeroPad(params.vset.toString(2), 8);
    var delay = zeroPad(params.delay.toString(2), 6);
    
    return delay + vset + sel + regulator + opcodeMap['REG_WRITE_VOUT_IMM'];
}

var emitRegWriteVctrlImm = function(params) {
    
    var regulator = zeroPad(params.regulator.toString(2), 4);
    var delayMode = zeroPad(params.delayMode.toString(2), 2);
    var vctrl = zeroPad(params.vctrl.toString(2), 8);
    var mask = zeroPad(params.mask.toString(2), 8);
    var delay = zeroPad(params.delay.toString(2), 6);
    
    return delay + mask + vctrl + delayMode + regulator + opcodeMap['REG_WRITE_VCTRL_IMM'];
}

var emitRegWriteMaskSreg = function(params) {
    
    var addr = zeroPad(params.addr.toString(2), 10);
    var sreg = zeroPad(params.sreg.toString(2), 2);
    var mask = zeroPad(params.mask.toString(2), 8);
    
    return mask + sreg + addr + opcodeMap['REG_WRITE_MASK_SREG'];
}

var emitSregReadReg = function(params) {
    
    var addr = zeroPad(params.addr.toString(2), 10);
    var sreg = zeroPad(params.sreg.toString(2), 2);
    
    return sreg + addr + opcodeMap['SREG_READ_REG'];
}


var emitWait = function(params) {
    
    var condition = zeroPad(params.condition.toString(2), 6);
    var conditionType = zeroPad(params.conditionType.toString(2), 2);
    var timeout = zeroPad(params.timeout.toString(2), 6);
    var skipCount = zeroPad(params.skipCount.toString(2), 6);
    
    return skipCount + timeout + conditionType + condition + opcodeMap['WAIT'];
}

var emitDelayImm = function(params) {
    
    var delay = zeroPad(params.delay.toString(2), 8);
    
    return delay + opcodeMap['DELAY_IMM'];
}
var emitDelaySreg = function(params) {
    
    var sreg = zeroPad(params.sreg.toString(2), 2);
    
    return sreg + opcodeMap['DELAY_SREG'];
}

var emitTrigSet = function(params) {
    
    var trigId = zeroPad(params.trigId.toString(2), 5);
    var trigSel = zeroPad(params.trigSel.toString(2), 6);
    var trigType = zeroPad(params.trigType.toString(2), 2);
    var imm = params.imm.toString(2);
    var ext = params.ext.toString(2);
    
    var destination = zeroPad(params.destination.toString(2), 13);
    
    if(params.instructionSetVersion >= 2) {
        if(ext === '1') {
            return destination + ext + imm + trigType + trigSel + trigId + opcodeMap['TRIG_SET'];
        } else {
            var reentrant = params.reentrant.toString(2);
            destination = destination.substring(0, 12);
            return destination + reentrant + ext + imm + trigType + trigSel + trigId + opcodeMap['TRIG_SET']
        }
    } else {
        return destination + ext + imm + trigType + trigSel + trigId + opcodeMap['TRIG_SET'];
    }
}

var emitTrigMask = function(params) {
    
    var trigMask = zeroPad(params.trigMask.toString(2), 28);
    
    return trigMask + opcodeMap['TRIG_MASK'];
}

var emitRegWriteMaskPage0Imm = function(params) {
    
    var addr = zeroPad(params.addr.toString(2), 8);
    var data = zeroPad(params.data.toString(2), 8);
    var mask = zeroPad(params.mask.toString(2), 8);
    
    return mask + data + addr + opcodeMap['REG_WRITE_MASK_PAGE0_IMM'];
}

var emitRegWriteBitPage0Imm = function(params) {
    
    var addr = zeroPad(params.addr.toString(2), 8);
    var data = params.data.toString(2);
    var bitSel = zeroPad(params.bitSel.toString(2), 3);
    
    return bitSel + data + addr + opcodeMap['REG_WRITE_BIT_PAGE0_IMM'];
}

var emitSregWriteImm = function(params) {
    
    var sreg = zeroPad(params.sreg.toString(2), 2);
    var data = zeroPad(params.data.toString(2), 8);
    
    return data + sreg + opcodeMap['SREG_WRITE_IMM'];
}

var emitRegWriteWinPage0Imm = function(params) {
    
    var addr = zeroPad(params.addr.toString(2), 8);
    var data = zeroPad(params.data.toString(2), 5);
    var mask = zeroPad(params.mask.toString(2), 5);
    var shift = zeroPad(params.shift.toString(2), 2);
    
    return shift + mask + data + addr + opcodeMap['REG_WRITE_WIN_PAGE0_IMM'];
}

var emitLiteral = function(params) {
    return params.binary;
}

var parsedToString = function(parsedInstruction) {
    if(parsedInstruction.command === 'REG_WRITE_MASK_PAGE0_IMM') {

        var params = parsedInstruction.params;
        var mask = zeroPad(params.mask.toString(16).toUpperCase(), 2);
        var data = zeroPad(params.data.toString(16).toUpperCase(), 2);
        var addr = zeroPad(params.addr.toString(16).toUpperCase(), 2);
        
        return "REG_WRITE_MASK_PAGE0_IMM ADDR=0x" + addr + " DATA=0x" + data + " MASK=0x" + mask;
    } else if(parsedInstruction.command === 'REG_WRITE_IMM') { // REG_WRITE_IMM
        var params = parsedInstruction.params;
        data = zeroPad(params.data.toString(16).toUpperCase(), 2);
        addr = params.addr;
        if(addr >= 0x300) {
            addr += 0x100;
        }
        addr = zeroPad(addr.toString(16).toUpperCase(), 3);
        
        return "REG_WRITE_IMM ADDR=0x" + addr.toString(16) + " DATA=0x" + data.toString(16);
    } else if(parsedInstruction.command === 'REG_WRITE_MASK_IMM') { // REG_WRITE_MASK_IMM
        
        var params = parsedInstruction.params;
        mask = zeroPad(params.mask.toString(16).toUpperCase(), 2);
        data = zeroPad(params.data.toString(16).toUpperCase(), 2);
        addr = params.addr;
        if(addr >= 0x300) {
            addr += 0x100;
        }
        addr = zeroPad(addr.toString(16).toUpperCase(), 3);
        return "REG_WRITE_MASK_IMM ADDR=0x" + addr.toString(16) + " DATA=0x" + data.toString(16) + " MASK=0x" + mask;
    } else if(parsedInstruction.command === 'REG_WRITE_VOUT_IMM') { // REG_WRITE_VOUT_IMM
        
        var params = parsedInstruction.params;
        delay = params.delay;
        vout = params.vset;
        selBin = zeroPad(params.sel.toString(2), 2);
        sel = reverseVoutSelMap[selBin];
        regulator = reverseRegulatorIdMap[params.regulator];
        
        // Convert vout binary to voltage
        var voutVolts;
        var bypass;
        if(regulator.substring(0,4) === "BUCK") {
            if(vout <= 0x0E) {
                voutVolts = vout * 0.02 + 0.3;
            } else if(vout <= 0x72) {
                voutVolts = vout * 0.005 + 0.525;
            } else if(vout <= 0xAA) {
                voutVolts = vout * 0.01 - 0.05;
            } else {
                voutVolts = vout * 0.02 - 1.76;
            }
        } else { // LDO
            if(regulator === "LDO4") {
                voutVolts = vout * 0.025 + 0.4;
                bypass = "0";
            } else {
                if((vout & 0x80) === 0x80) {
                    bypass = "1";
                    vout = vout & 0x7F;
                } else {
                    bypass = "0";
                }
                
                voutVolts = (vout >> 1) * 0.05 + 0.4;
            }
        }
        
        voutVolts = Math.round(voutVolts * 1000) / 1000; // Round to 3 decimal places
        
        var retString = "REG_WRITE_VOUT_IMM DELAY=" + delay + " VOUT=" + voutVolts + "V SEL=" + sel + " REGULATOR=" + regulator;
        if(regulator.substring(0,3) === "LDO") {
            retString += " BYPASS=" + bypass;
        }
        
        return retString;
    } else if(parsedInstruction.command === 'REG_WRITE_VCTRL_IMM') { // REG_WRITE_VCTRL_IMM
        var params = parsedInstruction.params;
        delay = params.delay;
        mask = zeroPad(params.mask.toString(16).toUpperCase(), 2);
        vctrl = zeroPad(params.vctrl.toString(16).toUpperCase(), 2);
        delayModeBin = zeroPad(params.delayMode.toString(2), 2);
        delayMode = reverseDelayModeMap[delayModeBin];
        regulator = reverseRegulatorIdMap[params.regulator];
        
        return "REG_WRITE_VCTRL_IMM DELAY=" + delay + " MASK=0x" + mask + " VCTRL=0x" + vctrl + " DELAY_MODE=" + delayMode + " REGULATOR=" + regulator;
    } else if(parsedInstruction.command === 'REG_WRITE_MASK_SREG') { // REG_WRITE_MASK_SREG
        var params = parsedInstruction.params;
        mask = zeroPad(params.mask.toString(16).toUpperCase(), 2);
        sreg = "R" + params.sreg.toString(10);
        addr = zeroPad(params.addr.toString(16).toUpperCase(), 3);
        
        return "REG_WRITE_MASK_SREG REG=" + sreg + " ADDR=0x" + addr + " MASK=0x" + mask;
    } else if(parsedInstruction.command === 'SREG_READ_REG') { // SREG_READ_REG
        var params = parsedInstruction.params;
        sreg = "R" + params.sreg;
        addr = zeroPad(params.addr.toString(16).toUpperCase(), 3);
        
        return "SREG_READ_REG REG=" + sreg + " ADDR=0x" + addr;
    } else if(parsedInstruction.command === 'WAIT') { // WAIT
        var params = parsedInstruction.params;
        skipCount = params.skipCount;
        timeout = params.timeout;
        conditionTypeBin = zeroPad(params.conditionType.toString(2), 2);
        conditionType = reverseTriggerTypeMap[conditionTypeBin];
        conditionSelBin = zeroPad(params.condition.toString(2), 6);
        conditionSel = reverseConditionSelectMap[conditionSelBin];
        
        return "WAIT DEST=" + params.destLabel + " TIMEOUT=" + timeout + " COND=" + conditionSel + " TYPE=" + conditionType;
    } else if(parsedInstruction.command === 'DELAY_IMM') { // DELAY_IMM
        return "DELAY_IMM " + parsedInstruction.params.delay;
    } else if(parsedInstruction.command === 'DELAY_SREG') { // DELAY_SREG
        return "DELAY_SREG R" + parsedInstruction.params.sreg;
    } else if(parsedInstruction.command === 'TRIG_SET') { // TRIG_SET
        var params = parsedInstruction.params; 
        reentrant = params.reentrant;
        ext = params.ext
        imm = params.imm;
        trigTypeBin = zeroPad(params.trigType.toString(2), 2);
        trigType = reverseTriggerTypeMap[trigTypeBin];
        trigSelBin = zeroPad(params.trigSel.toString(2), 6);
        trigSel = reverseTriggerSelectMap[trigSelBin];
        trigId = params.trigId;

        return "TRIG_SET DEST=" + params.destLabel + " ID=" + trigId + " SEL=" + trigSel + " TYPE=" + trigType + " IMM=" + imm + " EXT=" + ext + " REENTRANT=" + reentrant;
    } else if(parsedInstruction.command === 'TRIG_MASK') { // TRIG_MASK
        return "TRIG_MASK 0x" + parsedInstruction.params.trigMask.toString(16).toUpperCase();
    } else if(parsedInstruction.command === 'END') { // END
        return 'END';
    } else if(parsedInstruction.command === 'REG_WRITE_BIT_PAGE0_IMM') { // REG_WRITE_BIT_PAGE0_IMM
        var params = parsedInstruction.params;         
        var bitSel = params.bitSel;
        var data = params.data;
        var addr = zeroPad(params.addr.toString(16).toUpperCase(), 2);
        
        return "REG_WRITE_BIT_PAGE0_IMM ADDR=0x" + addr + " BIT=" + bitSel + " DATA=0x" + data;
    } else if(parsedInstruction.command === 'REG_WRITE_WIN_PAGE0_IMM') { // REG_WRITE_WIN_PAGE0_IMM
        var params = parsedInstruction.params;
        var shift = params.shift;
        var mask = zeroPad(params.mask.toString(16).toUpperCase(), 2);
        var data = zeroPad(params.data.toString(16).toUpperCase(), 2);
        var addr = zeroPad(params.addr.toString(16).toUpperCase(), 2);
        
        return "REG_WRITE_WIN_PAGE0_IMM ADDR=0x" + addr + " DATA=0x" + data + " MASK=0x" + mask + " SHIFT=" + shift;
    } else if(parsedInstruction.command === 'SREG_WRITE_IMM') { // SREG_WRITE_IMM
        var params = parsedInstruction.params;
        var data = zeroPad(params.data.toString(16).toUpperCase(), 2);
        var sreg = "R" + params.sreg;
        
        return "SREG_WRITE_IMM REG=" + sreg + " DATA=0x" + data;
    }
}

var optimizeProgram = function(programParsed, optimizeConfig, instructionSetVersion) {
    
    if(optimizeConfig.combineDelays) {
        programParsed = combineDelays(programParsed);
    }
    
    var newProgram = [];
    for(var i=0; i<programParsed.length; i++) {
        if(optimizeConfig.optRegWrite && programParsed[i].command == "REG_WRITE_MASK_IMM") {
            newProgram = newProgram.concat(optimizeRegWriteMaskImm(programParsed[i], instructionSetVersion));
        } else if(optimizeConfig.optRegWriteVout && programParsed[i].command == "REG_WRITE_VOUT_IMM") {
            newProgram = newProgram.concat(programParsed[i] = optimizeRegWriteVoutImm(programParsed[i]));
        } else if(optimizeConfig.optRegWriteVctrl && programParsed[i].command == "REG_WRITE_VCTRL_IMM") {
            newProgram = newProgram.concat(optimizeRegWriteVctrlImm(programParsed[i], instructionSetVersion));
        } else {
            newProgram.push(programParsed[i]);
        }
    }
     
    return newProgram;
}

var getNumMaskBits = function(maskBin) {
    
    var numMaskBits;
    var maskMatch = maskBin.match(/1/g); // Count number of '1' in mask field
    if(maskMatch === null) {
        numMaskBits = 0;
    } else {
        numMaskBits = maskMatch.length;
    }
    
    return numMaskBits;
}

var optimizeRegWriteMaskImm = function(parsedInstruction, instructionSetVersion) {
    
    var addr = parsedInstruction.params.addr;
    var mask = parsedInstruction.params.mask;
    var data = parsedInstruction.params.data;
    
    var maskBin = zeroPad(mask.toString(2), 8);
    var numMaskBits = getNumMaskBits(maskBin);
    
    var msbUnmasked = 7-maskBin.indexOf('0');
    var lsbUnmasked = 7-maskBin.lastIndexOf('0');
    var newInstruction;
    
    if(instructionSetVersion >=2 && numMaskBits === 7 && addr <= 0xFF) {
        var bit = (data & ~mask) === 0 ? 0 : 1;
        var params = {bitSel: msbUnmasked, data: bit, addr: addr};
        newInstruction = {command: "REG_WRITE_BIT_PAGE0_IMM", emitFunction: emitRegWriteBitPage0Imm, params: params, labels: parsedInstruction.labels};
    } else if(mask === 0) {
        var params = {addr: addr, data: data};
        newInstruction = {command: "REG_WRITE_IMM", emitFunction: emitRegWriteImm, params: params, labels: parsedInstruction.labels};
    } else if(instructionSetVersion >= 2 && msbUnmasked-lsbUnmasked <= 4 && addr <= 0xFF) {
        var windowMask;
        var windowData;
        var shift;
        var dataBin = zeroPad(data.toString(2), 8);
        if(msbUnmasked <= 4) {
            shift = 0;
            windowMask = maskBin.substring(3, 8);
            windowData = dataBin.substring(3, 8);
        } else {
            shift = msbUnmasked-4;
            windowMask = maskBin.substring(7-msbUnmasked, 7-msbUnmasked+5);
            windowData = dataBin.substring(7-msbUnmasked, 7-msbUnmasked+5);
        }
        
        var params = {shift: shift, mask: parseInt(windowMask, 2), data: parseInt(windowData, 2), addr: addr};
        newInstruction = {command: "REG_WRITE_WIN_PAGE0_IMM", emitFunction: emitRegWriteWinPage0Imm, params: params, labels: parsedInstruction.labels};
    } else if(instructionSetVersion >= 2 && addr <= 0xFF) {
        var params = {addr: addr, data: data, mask: mask};
        newInstruction = {command: "REG_WRITE_MASK_PAGE0_IMM", emitFunction: emitRegWriteMaskPage0Imm, params: params, labels: parsedInstruction.labels}; 
    } else {
        newInstruction = parsedInstruction;
    }
    
    return [newInstruction];
}

var optimizeRegWriteVctrlImm = function(parsedInstruction, instructionSetVersion) {
    
    var regulator = parsedInstruction.params.regulator;
    var vctrl = parsedInstruction.params.vctrl;
    var mask = parsedInstruction.params.mask;
    var delay = parsedInstruction.params.delay;
    var delayMode = parsedInstruction.params.delayMode;
    
    var newInstruction;
    
    if(delay === 0) {
        var regulatorName = reverseRegulatorIdMap[regulator];
        var registerName = regulatorName + "_CTRL";
        var registerAddress = registerAddressMap[registerName];
        
        var params = {addr: registerAddress, data: vctrl, mask: mask};
        newInstruction = {command: "REG_WRITE_MASK_IMM", emitFunction: emitRegWriteMaskImm, params: params, labels: parsedInstruction.labels};
        return optimizeRegWriteMaskImm(newInstruction, instructionSetVersion);
        
    } else {
        var maskBin = zeroPad(mask.toString(2), 8);
        var numMaskBits = getNumMaskBits(maskBin);
        if(instructionSetVersion >= 2 && numMaskBits === 7) {
            // DELAY_IMM + REG_WRITE_BIT_PAGE0_IMM is shorter than REG_WRITE_VCTRL_IMM
            var delayParams = {delay: delay};
            var newDelayImm = {command: "DELAY_IMM", emitFunction: emitDelayImm, params: delayParams, labels: parsedInstruction.labels};
            
            var bitSel = 7-maskBin.indexOf('0');
            var bit = (vctrl & ~mask) === 0 ? 0 : 1;
            var regulatorName = reverseRegulatorIdMap[regulator];
            var registerName = regulatorName + "_CTRL";
            var addr = registerAddressMap[registerName];
            var bitWriteParams = {bitSel: bitSel, data: bit, addr: addr};
            var newBitWrite = {command: "REG_WRITE_BIT_PAGE0_IMM", emitFunction: emitRegWriteBitPage0Imm, params: bitWriteParams, labels: parsedInstruction.labels};
            
            return [newDelayImm, newBitWrite];
        } else {
            newInstruction = parsedInstruction;
        }
    }
    
    return [newInstruction];
}

var optimizeRegWriteVoutImm = function(parsedInstruction) {
    
    var regulator = parsedInstruction.params.regulator;
    var regulatorName = reverseRegulatorIdMap[regulator];
    var sel = parsedInstruction.params.sel;
    var vset = parsedInstruction.params.vset;
    var delay = parsedInstruction.params.delay;
    
    var newInstruction;
    
    if(delay === 0 && (regulatorName.indexOf("BUCK") !== -1) && (sel === 0 || sel === 1)) {
        var registerName;
        if(sel === 0) { // Writing to VOUT1
            registerName = regulatorName + "_VOUT_1";
        } else {        // Writing to VOUT2
            registerName = regulatorName + "_VOUT_2"; 
        }
        
        var registerAddress = registerAddressMap[registerName];
        var params = {addr: registerAddress, data: vset};
        newInstruction = {command: "REG_WRITE_IMM", emitFunction: emitRegWriteImm, params: params, labels: parsedInstruction.labels};
    } else if(delay === 0 && regulatorName.indexOf("LDO") !== -1) {
        var registerName = regulatorName + "_VOUT";
        var registerAddress = registerAddressMap[registerName];
        var params = {addr: registerAddress, data: vset};
        newInstruction = {command: "REG_WRITE_IMM", emitFunction: emitRegWriteImm, params: params, labels: parsedInstruction.labels};
    } else {
        newInstruction =  parsedInstruction;
    }
    
    return [newInstruction];
}

 
var combineDelays = function(programParsed) {
    newProgram = [];
    
    var i = 0;
    while(i < programParsed.length) {
        // Last instruction cannot be optimized, so it always goes in the program
        if(i+1 == programParsed.length) {
            newProgram.push(programParsed[i]);
            break;
        }
        
        if(programParsed[i].command == "DELAY_IMM") {
            var labels = programParsed[i].labels;
            var newDelay = programParsed[i].params.delay;
            
            // Combines all consecutive delays which don't have a label, and that fit in a single delay instruction
            while(i+1 < programParsed.length && programParsed[i+1].command == "DELAY_IMM" && programParsed[i+1].labels.length == 0) {
                if(programParsed[i+1].params.delay + newDelay > 256) {
                    // delay is too big to fit in the instruction
                    break;
                }
                i += 1;
                newDelay += programParsed[i].params.delay;
            }
            
            var combineRegWrite = false;
            // Combine with REG_WRITE_VCTRL_IMM or REG_WRITE_VOUT_IMM if it's the next instruction
            if(i+1 < programParsed.length) {
                var nextCommand = programParsed[i+1].command
                if(nextCommand == "REG_WRITE_VCTRL_IMM" || nextCommand == "REG_WRITE_VOUT_IMM") {
                    if(newDelay + programParsed[i+1].params.delay < 64) {                       
                        i += 1;
                        
                        // We can alias the parsed program, because we're dropping it in this function anyway
                        var newInstruction = programParsed[i];
                        newDelay += programParsed[i].params.delay;
                        newInstruction.params.delay = newDelay;
                        
                        newInstruction.labels = labels;
                        
                        newProgram.push(newInstruction);
                        combineRegWrite = true;
                    }
                }
            }
            
            // Cannot combine with next instruction, so just combine the DELAY_IMM commands
            if(!combineRegWrite) {
                var params = {delay: newDelay};
                var newInstruction = {command: "DELAY_IMM", emitFunction: emitDelayImm, params: params};
                newInstruction.labels = labels;
                newProgram.push(newInstruction);
            }
            
        } else {
            // Cannot optimize
            newProgram.push(programParsed[i]);
        }
        
        i += 1;
    }
    
    return newProgram;
}